@@ -55,14 +55,16 @@ inline void appSuf(string &toadd, unsigned int i) {
 }
 
 // global lock for all decission critical changes
-pthread_mutex_t chlock = PTHREAD_MUTEX_INITIALIZER;
+pthread_mutex_t nameres_lock = PTHREAD_MUTEX_INITIALIZER;
 
 class lf /*: public lockable*/ {
    public:
       lf(const char *path) { 
          csize=ncsize; // default for writters, can be changed by existing value from the first chunk
-         cache_fd=-1;
-         cache_pos=-1;
+         ccount=0;
+         cached_fd=-1;
+         cached_local_off=-1;
+         cached_chunk=-1;
 
          struct stat stbuf;
          if(!path)
@@ -70,7 +72,7 @@ class lf /*: public lockable*/ {
          virtname=fromdir+path;
 
          if(0==lstat(virtname.c_str(), &stdata)) { // unchunked file, nothing to do here
-            fflags.resize(1);
+            ccount=1;
             return;
          }
 
@@ -85,40 +87,41 @@ class lf /*: public lockable*/ {
          if(stdata.st_size)
             csize=stdata.st_size; // chunk size to be saved as default for this object
          stdata.st_size=0; // not double-counting the first chunk
-         for(unsigned int i=0;i<=676;i++) {
-            if(!lstat(makeChunkName(i).c_str(), &stbuf)) {
+         for(unsigned int i=0;i<676;i++) {
+            if(0==lstat(makeChunkName(i).c_str(), &stbuf)) {
                if( (uint64_t)stbuf.st_size > csize) {
-                  // FIXME, report error and mark invalid
-                  fflags.clear();
-               }
-               else if( (uint64_t)stbuf.st_size < csize) {
-                  // last chunk
-                  fflags.resize(i+1);
-                  stdata.st_size+=stbuf.st_size;
+                  ccount=-1;
                   return;
+                  // FIXME, report error and mark invalid
+                  //compflags.clear();
                }
-               else
+               else {
+                  ccount=i+1;
                   stdata.st_size+=stbuf.st_size;
+                  if( (uint64_t)stbuf.st_size < csize)
+                     // last chunk
+                     return;
+               }
             }
             else {
-               fflags.resize(i);
+               ccount=i; // one less failed i -> ok
                return;
             }
          }
       };
 
       inline string getCname(unsigned int i) {
-         //setLockGuard;
-         if(i>=fflags.size())
-            return "";
-         if(fflags.size()<2)
+         // FIXME: why? 
+         // if(i>=compflags.size())
+         //   return "";
+         if(ccount<2) // same name
             return virtname;
          return makeChunkName(i);
       }
 
       ~lf() { 
-         if(cache_fd>=0)
-            close(cache_fd);
+         if(cached_fd>=0)
+            close(cached_fd);
       }
 
       void getFstat(struct stat *stbuf) {
@@ -126,13 +129,13 @@ class lf /*: public lockable*/ {
       }
 
       bool exists() { 
-         return (fflags.size()); 
+         return (ccount>0);
       }
 
       bool prep_open(int flags) {
-         if(fflags.empty()) {
+         if(0==ccount) {
             if((flags&O_CREAT))
-               fflags.resize(1); // force the use the virtname for first chunk
+               ccount=1; // force the use of the virtname for first chunk
             else {
                errno=ENOENT;
                return false;
@@ -144,109 +147,129 @@ class lf /*: public lockable*/ {
 
       int getReaderAt(off_t offset) {
 
-         unsigned int chunk = offset/csize;
-         int64_t coff=offset%csize;
+         unsigned int local_chunk = offset/csize;
+         int64_t local_off=offset%csize;
 
          // try it either the easy or the hard way
-         if( ! (cache_fd>=0 && chunk==(unsigned int)cache_chunk && coff == (unsigned int) cache_pos))
+         if( ! (cached_fd>=0 && local_chunk==(unsigned int)cached_chunk && local_off == (unsigned int) cached_local_off))
          { // don't rely on anything, make it safe
-            if(cache_fd>=0)
-               close(cache_fd);
-            // getCname resolution must be valid when open happens
-            pthread_mutex_lock(&chlock);
-            cache_fd = open(getCname(chunk).c_str(), flags);
-            pthread_mutex_unlock(&chlock);
-            // double-check locking optimisation, safe enough
-            if(cache_fd<0)
-               return cache_fd;
+            if(cached_fd>=0)
+               close(cached_fd);
+            
+            pthread_mutex_lock(&nameres_lock); // getCname resolution must be valid when open happens
+            cached_fd = open(getCname(local_chunk).c_str(), flags); // abuse cached_fd here, does not matter
+            pthread_mutex_unlock(&nameres_lock);
+            if(cached_fd<0)
+               return cached_fd;
+            cached_chunk=local_chunk;
 
-            if(coff!=lseek(cache_fd, coff, SEEK_SET))
+            if(local_off!=lseek(cached_fd, local_off, SEEK_SET))
                return -1;
+            cached_local_off=local_off;
 
          }
 
-         return cache_fd;
+         return cached_fd;
       }
       void moveOffset(off_t offset) {
-         cache_pos=offset%csize;
+         cached_local_off=offset%csize;
+      }
+
+      inline void pad_before(unsigned int current_chunk) {
+         // make sure the previous have the right size
+         unsigned int i = 0;
+         if(ccount>=2)
+            i = ccount-2;
+         for(;i<current_chunk;i++)
+            truncate(getCname(i).c_str(), csize);
+      }
+
+      inline int unlink_after(unsigned int current_chunk) {
+
+         int ret(0);
+
+         // drop the trailing
+         for(unsigned int i=current_chunk+1; i<ccount; i++)
+            if(0!=unlink(getCname(i).c_str()))
+               if(ret==0) // remember only the first error
+                  ret=errno;
+
+         return ret;
       }
 
       int trunc(off_t size) {
-         unsigned int chunks = size/csize;
-         int64_t coff=size%csize;
+         unsigned int current_chunk = size/csize;
+         int64_t last_off=size%csize;
+
          // this may suck with slowly deleting filesystems
-         pthread_mutex_lock(&chlock);
-         // pad preceeding, truncate the last one, drop the remaining rest
-         unsigned int i;
-         for(i=0;i<chunks;i++) {
-            if(!fflags[i]) {
-               truncate(getCname(i).c_str(), csize);
-               fflags[i]=true;
-            }
-         }
+         pthread_mutex_lock(&nameres_lock);
+
+         pad_before(current_chunk);
+         unlink_after(current_chunk);
+         ccount=current_chunk+1;
+
          //cerr << "truncating " << getCname(i).c_str() << " to " <<coff <<endl;
-         truncate(getCname(i).c_str(), coff);
-         fflags[i]=false; // don't know for sure, make it so
-         for(i++;i<fflags.size();i++)
-            unlink(getCname(i).c_str());
-         fflags.resize(chunks+1); // update count
+         truncate(getCname(current_chunk).c_str(), last_off);
+
          stdata.st_size=size;
          // cerr << "neu: " << ccnt << " chunks, " << size << " bytes\n";
-         pthread_mutex_unlock(&chlock);
+         //
+         pthread_mutex_unlock(&nameres_lock);
          return 0;
       }
 
       int getWriterAt(off_t offset, size_t &rest) {
-         unsigned int chunk = offset/csize;
-         int64_t coff=offset%csize;
-         if( ! (cache_fd>=0 && chunk==(unsigned int)cache_chunk && coff == (unsigned int) cache_pos))
+         unsigned int current_chunk = offset/csize;
+         if(current_chunk>=26*26) {
+            errno=ENOSPC;
+            return -1;
+         }
+         int64_t local_off=offset%csize;
+         if( ! (cached_fd>=0 && current_chunk==(unsigned int)cached_chunk && local_off == (unsigned int) cached_local_off))
          {
-            if(cache_fd>=0)
-               close(cache_fd);
-            // getCname resolution must be valid when open happens
-            pthread_mutex_lock(&chlock);
-            if(coff==0 && chunk == 1 && fflags.size()==1) {
+            if(cached_fd>=0)
+               close(cached_fd);
+
+            pthread_mutex_lock(&nameres_lock);
+            if(local_off==0 && current_chunk == 1 && ccount==1)
+            {
                 // time to rename, must be locked here
                 rename(virtname.c_str(), (virtname+".aa").c_str()); // FIXME: error handling
             }
-            if(chunk>=fflags.size())
-               fflags.resize(chunk+1);
 
-            // make sure the previous have the right size.
-            if(chunk>0) {
-               for(unsigned int i=0;i<chunk;i++) {
-                  if(!fflags[i]) {
-                     truncate(getCname(i).c_str(), csize);
-                     fflags[i]=true;
-                  }
-               }
-            }
-            
+            pad_before(current_chunk);
+
+            if(current_chunk > ccount-1)
+               ccount=current_chunk+1;
+
             //cerr << "> " << getCname(chunk).c_str() << " flags: " << (flags|O_CREAT) <<endl;
-            cache_fd = open(getCname(chunk).c_str(), O_CREAT|O_RDWR, S_IRUSR|S_IWUSR|S_IRGRP);
-            pthread_mutex_unlock(&chlock);
-            if(cache_fd<0)
+            cached_fd = open(getCname(current_chunk).c_str(), O_CREAT|O_RDWR, S_IRUSR|S_IWUSR|S_IRGRP);
+            pthread_mutex_unlock(&nameres_lock);
+            if(cached_fd<0)
                return -1;
-            cache_pos=coff;
-            lseek(cache_fd, coff, SEEK_SET);
+            cached_chunk=current_chunk;
+            if(local_off!=lseek(cached_fd, local_off, SEEK_SET))
+               return -1;
+            cached_local_off=local_off;
          }
-         if(rest > csize-cache_pos)
-            rest=csize-cache_pos;
-         return cache_fd;
+         if(rest > csize-cached_local_off)
+            rest=csize-cached_local_off;
+         return cached_fd;
       }
 
    private:
 
          // signed types, negative values for invalid
-      int cache_fd;
-      int cache_chunk;
-      int64_t cache_pos;
-      int flags;
+      int cached_fd;
+      int cached_chunk;
+      int64_t cached_local_off;
+
+      int ccount;
+      uint64_t csize;
 
       string virtname;
-      uint64_t csize;
       struct stat stdata;
-      vector<bool> fflags; // flag bits for "full files" to speedup padding (which is O(n^2) * syscall time) ; it also serves as chunk counter
+      int flags;
 
       inline string makeChunkName(unsigned int i) {
          string ret=virtname;
@@ -362,8 +385,8 @@ proc_next_no_inc:
             tmpit=it;
             it++;
             smap.erase(tmpit);
+            // check next name
             incName(toskip);
-            // process next without increasing
             goto proc_next_no_inc;
          }
       }
@@ -382,23 +405,6 @@ proc_next_no_inc:
          break;
    }
 
-
-   /*
-   DIR *dp = get_dirp(fi);
-   struct dirent *de;
-
-   (void) path;
-   seekdir(dp, offset);
-   while ((de = readdir(dp)) != NULL) {
-      struct stat st;
-      memset(&st, 0, sizeof(st));
-      st.st_ino = de->d_ino;
-      st.st_mode = de->d_type << 12;
-      if (filler(buf, de->d_name, &st, telldir(dp)))
-         break;
-   }
-
-   */
    return 0;
 }
 
@@ -446,13 +452,7 @@ static int fulfs_unlink(const char *path)
    int err(0);
    int i(0);
    lf p(path);
-   while(true) {
-      string n=p.getCname(i++);
-      if(n.empty())
-         return (err?-errno:0);
-      // summarize all -1
-      err+=unlink(n.c_str());
-   }
+   return -p.unlink_after(-1);
 }
 
 static int fulfs_rmdir(const char *path)
@@ -706,23 +706,23 @@ static int fulfs_fsync(const char *path, int isdatasync,
 }
 
 
+struct fuse_operations fulfs_oper;
 
-void FuseUsage() {
-   cerr << "fulfs rootDir mountPoint [chunkSize] [FUSE Mount Options]\n"
+void _ExitUsage() {
+   cerr << "USAGE: fulfs rootDir mountPoint [chunkSize] [FUSE Mount Options]\n"
       << "The chunkSize argument specifies size for fresh files\n"
       << "(default 2GiB-1), suffixes for units are allowed, eg. 700M\n"
       << "valid FUSE Mount Options follow:\n";
-    int argc = 2;
     const char *argv[] = {"...", "-h"};
-    fuse_main( argc, const_cast<char**>(argv), (fuse_operations*)NULL);
+    fuse_main( 2, const_cast<char**>(argv), &fulfs_oper);
+    exit(EXIT_FAILURE);
 }
 
-#define barf(x) { cerr << "ERROR: " << x <<endl; FuseUsage(); exit(EXIT_FAILURE); }
+#define barf(x) { cerr <<endl << "ERROR: " << x <<endl<<endl; _ExitUsage(); }
 
 int main(int argc, char *argv[])
 {
 
-   struct fuse_operations fulfs_oper;
    memset(&fulfs_oper, 0, sizeof(fulfs_oper));
 
    fulfs_oper.getattr	= fulfs_getattr;
@@ -755,7 +755,7 @@ int main(int argc, char *argv[])
    umask(0);
 
    if(argc<3)
-      barf("\nNeeds a source and a target directory, see --help.\n");
+      barf("Needs a source and a target directory, see --help.");
 
    struct stat stbuf;
    if(stat(argv[1], &stbuf) || !S_ISDIR(stbuf.st_mode))
@@ -765,13 +765,13 @@ int main(int argc, char *argv[])
    if(realpath(argv[1], buf))
       fromdir=buf;
    else {
-      cerr << "\nInvalid source directory, " << argv[1] <<endl <<endl;
+      cerr << "Invalid source directory, " << argv[1] <<endl;
       exit(EXIT_FAILURE);
    }
    delete [] buf;
 
    if(stat(argv[2], &stbuf) || !S_ISDIR(stbuf.st_mode))
-      barf(endl<< argv[2] << " is not a directory.\n\n");
+      barf(endl<< argv[2] << " is not a directory.");
 
    todir=argv[2];
 
